''' ----------------------   Libraries importation   ---------------------- '''

# Librairies importation
import pandas as pd
import numpy as np
import datetime as dt
import random
from calendar import monthrange

''' ------------------   Data definition and importation   --------------------

INDICES
    i, j : Aircraft
    p, q : Phases (groups of services)
    s, t : Services
    v, k, l : Vehicles
    
AIRCRAFT DATA (AC)
    ID[i] : Aircraft i ID    
    type[i] : Aircraft i type (B734, A332, MD11 ...)
    TWa[i] : Lower bound of aircraft i expected operating time window (on block time)
        --> TWa[i] = STA[i] (Scheduled Time of Arrival of aircraft i) + 5 min of taxiing
    TWb[i] : Upper bound of aircraft i expected operating time window (off block time)
        --> TWb[i] = STD[i] (Scheduled Time of Departure of aircraft i) - 5 min of taxiing
    parking[i] : Aircraft i parking stand id
    IN[i] : Warehouse location where unloaded ULDs of aircraft i should be stored
    OUT[i] : Warehouse location where the ULDs that should be loaded in aircraft i are stored

ENVIRONMENT DATA (env)
    nb_loc : Number of locations available for vehicles (including AC parkings and warehouses)
    distances[x,y]: Distance (in m) between location x and location y

PHASES DATA (P)
    nb : Number of phases to perform on each aircraft
    a[p][i] : Earliest start time of phase p at aircraft i
    precedenceMatrix[p,q] : Binary matrix showing precedence relations between phases
        --> For each phase (line), a "1" means that the corresponding phase (column)
            should be performed before the considered operation (line).

SERVICES DATA (S)
    ID[s] : Service s ID
    phase[s] : Phase to which the service s is attached
    first[s] : True if s does not depend of the end of another service, False otherwise
    QGlobal[s] : True if the service served AC directly and is not involved in any goods transfer chain, 
                 False otherwise
    place1[s] : Place where the service s should start OR place of reference for vehicles selection
    place2[s] : Place where the service s should end
    loading[s] : True if the service s requires loading/unloading of a quantity that reduces the 
                 capacity of the vehicle concerned, False otherwise
    transpVehicle[s] : True if the goods carried by the vehicle involved in service s 
                       are other vehicles, False otherwise
    updatePosition[s] : True if service s requires the vehicle to move from its current position,
                        False otherwise
    ACquantity[s] : True if an aircraft-related quantity is to be determined for service s,
                    False otherwise
    simultaneity[s] : True if several vehicles of the same type can perform service s for 
                      the considered AC at the same time, False otherwise
    maxV[s] : Maximum number of different vehicles that can be selected to perform service s
    sameV[s] : True if (once maxV reached) no new vehicles can be selected to perform service s 
               (the selection should be done inside currentV[s][i]), False otherwise
    next[s] : List of service IDs that depends on service s
    q_init[s][i] : Demand quantity of aircraft i for service s 
              (0 if service s demand quantity is related to another service)
    S[p] = List of services of phase p for which first == True
    currentV[s][i] : List of current vehicles ID that have already been selected to perform service s on AC i
    currentNumber[s][i] : Current number of different vehicles performing service s on AC i

VEHICLES DATA (V)
    ID[k] : Vehicle k ID
    type[k] : Vehicle k type
    nb : Total number of vehicles considered 
    depot[k] : List of depots where vehicle k is allowed
    fpairings[k] : List of AC type that cannot be served by vehicle k
    services[k] : List of services that could be performed by vehicle k
    sequence[k] : List of service sequences that should be followed by vehicle k
    maxCap[k] : Maximum capacity of vehicle k
    speed[k] : Travel speed of vehicle k (in m/s)
    tbase[k] : Time required by vehicle k to complete the first part of any service it performs (in sec)
    c[k][s] : Constant time required by vehicle k for setup and post-operations of the sth service it could perform (in sec)
    cD[k][s] : Constant time required by vehicle k for setup and post-operations of the sth service it could perform at the depot (in sec)
    v[k][s] : Variable time required by vehicle k to deliver one unit of product in the sth service it could perform (in sec)
    vD[k][s] : Variable time required by vehicle k to replenish one unit of product related to the sth service it could perform (in sec)
    currentCap[k] : Current capacity of vehicle k
    currentService[k] : ID of the service currently performed by vehicle k (or that should be performed if involved in a service sequence)
        --> "any" if vehicle k is free to perform any of its related service (services)
    used[k] : True if vehicle k is currently used, False otherwise
    available[k] : Time at which vehicle k is again available
    ttmp[k]: Earliest start time of the service considered by vehicle k
    compatibleAC_init[k] : Initial list of AC compatible with vehicle k
    compatibleAC[k] : Current list of AC compatible with vehicle k
    currentPos[k] : Current position of vehicle k

SOLUTION DATA (sol)
    TripPosition[k] : List of positions (route) to follow by vehicle k in the solution
    TripSchedule[k] : List of time (schedule) that vehicle k has to respect in the solution
    EndTime[i][s] : End time of service s at aircraft i
    
--------------------------------------------------------------------------- '''


#####  -------------------  Aircraft characteristics  -------------------  #####

# Extract the data from the excel file containing your instances
doc_1 = pd.ExcelFile("your_instances_file.xlsx")
feuille_1 = pd.read_excel(doc_1, "one_instance")
rows_1 = len(feuille_1)

# Create aircraft data
class aircraft:
    ID = []
    type = []
    STA = []
    STD = []
    TWa = []
    TWb = []
    parking = []
    IN = []
    OUT = []
    
    # Fill empty lists with the extracted data
    for i in range(0, rows_1):
        ID.append(feuille_1.iloc[i,0])
        type.append(feuille_1.iloc[i,1])
        STA.append(feuille_1.iloc[i,2])
        #On block time = STA + 5 min of taxiing to the parking stand
        TWa.append(STA[i] + dt.timedelta(minutes=5))
        STD.append(feuille_1.iloc[i,5])
        #Off block time = STD - 5 min of taxiing to the runway
        TWb.append(STD[i] - dt.timedelta(minutes=5))
        parking.append(feuille_1.iloc[i,7])
        IN.append(feuille_1.iloc[i,8])
        OUT.append(feuille_1.iloc[i,9])
    
# Create a class object to use aircraft data in other data classes
AC = aircraft()


#####  --------------  Warehouses and parkings distances  --------------  #####

# Create environment data
class environment:
    nb_loc = "YOUR INSTANCE number of parking lots and depots considered"
    distances = np.empty((nb_loc,nb_loc))
    
    # Extract data from excel file containing the distances between the parking lots/depots of the the airport considered
    doc_2 = pd.ExcelFile("your_file.xlsx")
    feuille_2 = pd.read_excel(doc_2, "your_page")
    rows_2 = len(feuille_2)
    cols_2 = len(feuille_2.columns)
    
    # Fill the distances matrix with the extracted data
    for x in range(0, rows_2):
        for y in range (1, cols_2):
            distances[x,y-1] = feuille_2.iloc[x,y]


#####  ---------  Phases parameters and precedence constraints  ---------  #####

# Create phases data
class phase:
    nb = 7
    
    ## By default, phases' initial earliest start time is TWa but there exist an exception for group of service 1 (bringing empty dolly to the AC parking lot)
    a = [[] for p in range(0, nb)]
    for p in range(0, nb):
        for i in range(0, rows_1):
            if p == 1:
                a[p].append(AC.STA[i] - dt.timedelta(minutes=30))
            
            else:
                a[p].append(AC.TWa[i])
    
    
    ## Precedence constraints among phases
    precedenceMatrix = np.empty((nb, nb))
    
    # Extract data from excel file "precedence.xlsx"
    doc_3 = pd.read_excel("precedence.xlsx")
    rows_3 = len(doc_3)
    cols_3 = len(doc_3.columns)
    
    # Fill the precedence matrix with the extracted data
    for p in range(0, rows_3):
        for q in range (1, cols_3):
            precedenceMatrix[p,q-1] = doc_3.iloc[p,q]
    
    
    ## Authorize subsequent modifications of some variables
    def __init__(self):
        self.a = self.a

# Create a class object to use phases data in other data classes
P = phase()


#####  -------------------  Services characteristics  -------------------  #####

# Extract data from excel file containing the AC maximum capacity (toilet tank, fuel tank, water tank, number of ULDs) for each AC type considered
doc_6 = pd.read_excel("your_file.xlsx")
rows_6 = len(doc_6)

# Create services data
class service:
    ID = []
    phase = []
    first = []
    QGlobal = []
    place1 = []
    place2 = []
    loading = []
    transpVehicle = []
    updatePosition = []
    ACquantity = []
    simultaneity = []
    sameV = []
    maxV = []
    
    # Extract data from excel file "services.xlsx"
    doc_4 = pd.ExcelFile("services.xlsx")
    feuille_4 = pd.read_excel(doc_4, "services_data")
    rows_4 = len(feuille_4)
    
    # Fill empty lists with the extracted data
    for s in range(0, rows_4):
        ID.append(feuille_4.iloc[s,1])
        phase.append(feuille_4.iloc[s,3])
        first.append(feuille_4.iloc[s,4])
        QGlobal.append(feuille_4.iloc[s,6])
        place1.append(feuille_4.iloc[s,8])
        place2.append(feuille_4.iloc[s,9])
        loading.append(feuille_4.iloc[s,10])
        transpVehicle.append(feuille_4.iloc[s,11])
        updatePosition.append(feuille_4.iloc[s,12])
        ACquantity.append(feuille_4.iloc[s,13])
        simultaneity.append(feuille_4.iloc[s,7])
        sameV.append(feuille_4.iloc[s,14])
        maxV.append(feuille_4.iloc[s,15])
    
        
    ## Precedence constraints among services
    next = [[] for s in range(0, rows_4)]
    
    # Extract data from excel file "services.xlsx"
    feuille_5 = pd.read_excel(doc_4, "next")
    rows_5 = len(feuille_5)
    cols_5 = len(feuille_5.columns)
    
    # Fill the empty list with the extracted data
    for s in range(0, cols_5):
        for t in range (0, rows_5):
            if feuille_5.iloc[t,s] != -1 :
                next[s].append(feuille_5.iloc[t,s])
    
    
    ## Quantities to be served
    q_init = [[] for s in range(0, rows_4)]
    
    # Number of ULDs IN (AC to warehouse) should remain the same among the different services
    ULD_IN = []
    for i in range(0, rows_1):
        for j in range(0, rows_6):
                if AC.type[i] == doc_6.iloc[j,0]:
                    q_IN = doc_6.iloc[j,5]*0.9 #if you consider that each aircraft is filled at 90% of their capacity
                    #*round(random.uniform(0.8,1),2) if you want to add randomness
                    # Number of ULDs should be an integer
                    ULD_IN.append(round(q_IN))
    
    # Number of ULDs OUT (warehouse to AC) should remain the same among the different services
    ULD_OUT = []
    for i in range(0, rows_1):
        for j in range(0, rows_6):
            if AC.type[i] == doc_6.iloc[j,0]:
                q_OUT = doc_6.iloc[j,5]*0.9 #if you consider that each aircraft is filled at 90% of their capacity
                #*round(random.uniform(0.8,1),2) if you want to add randomness
                # Number of ULDs should be an integer
                ULD_OUT.append(round(q_OUT))
    
    # Fill demand quantity list based on aircraft type with the extracted data
    for s in range(0, rows_4):
        for i in range(0, rows_1):
            if ACquantity[s] == True:
                for j in range(0, rows_6):
                    if AC.type[i] == doc_6.iloc[j,0]:
                        if s == 0: #toilet emptying
                            q_init[s].append((doc_6.iloc[j,2])*0.75) #if you want to consider fixed quantity or *round(random.uniform(0.2,1),2))
                        elif s == 1: #fueling
                            q_init[s].append((doc_6.iloc[j,3])*0.6) #if you want to consider fixed quantity or *round(random.uniform(0.5,0.9),2))
                        elif s == 5: #potable water servicing
                            q_init[s].append((doc_6.iloc[j,4])*0.75) #if you want to consider fixed quantity or *round(random.uniform(0.2,1),2))
                        elif s == 6 or s == 8 or s == 20: #ULDs IN
                            q_init[s].append(ULD_IN[i])
                        elif s == 12 or s == 14 or s == 16: #ULDs OUT
                            q_init[s].append(ULD_OUT[i])
                        else: #Catering, safety inspection, cleaning
                            q_init[s].append(1) 
            else:
                q_init[s].append(0)                
    
    
    ## Phases' starting services
    S = [[] for p in range(0, P.nb)]
    for s in ID :
    	if first[s] == True: 
    		S[phase[s]].append(s)
    
        
    ## Vehicles used to perform the service
    currentNumber = np.zeros((rows_4, rows_1))
    currentV = [[] for s in range(0, rows_4)]
    for s in range(0, len(currentV)):
        for i in AC.ID:
            currentV[s].append([])
    

    ## Authorize subsequent modifications of some variables
    def __init__(self):
        # Create a copy of q_ to keep stored the initial quantity that had to be served
        self.q = self.q_init.copy()
        self.currentNumber = self.currentNumber.copy()
        self.currentV = self.currentV.copy()


# Create a class object to use services data in other data classes
S = service()

        
#####  ---------------------  Vehicles parameters  ---------------------  #####

# Create vehicles data
class vehicle:

    # Extract data from excel file "vehiclesAll.xlsx" or "vehiclesMin.xlsx" depending on the number of vehicles considered
    doc_7 = pd.ExcelFile("vehiclesMin.xlsx")
    MinResources = True #False if "vehiclesAll.xlsx" used
    feuille_7 = pd.read_excel(doc_7, "vehicles")
    rows_7 = len(feuille_7)
    feuille_8 = pd.read_excel(doc_7, "depot", header = None)
    rows_8 = len(feuille_8)
    cols_8 = len(feuille_8.columns)
    feuille_9 = pd.read_excel(doc_7, "v_services", header = None)
    rows_9 = len(feuille_9)
    cols_9 = len(feuille_9.columns)
    feuille_10 = pd.read_excel(doc_7, "f_pairings", header = None)
    rows_10 = len(feuille_10)
    cols_10 = len(feuille_10.columns)


    ## Vehicles creation
    ID = []
    type = []
    
    for k in range(0, rows_7):
        # If the vehicle considered is the AC or the warehouse IN, their number equal the nb of AC of the instances
        if feuille_7.iloc[k,0] == 18 or feuille_7.iloc[k,0] == 19:
            n = rows_1
        
        # For the other vehicle types, define the number depending on the case considered (all vehicles or minimum)
        else:
            
            # If we reduce the number of vehicles available
            if MinResources == True:
                # The number of dollies, HL, SL, and cargo tractors is also dependent of the instance size
                if feuille_7.iloc[k,0] == 14:
                    n = min(max(round(2.75*rows_1),70), 484)
                
                elif feuille_7.iloc[k,0] == 15:
                    n = min(max(round(0.125*rows_1),4), 38)
                    
                elif feuille_7.iloc[k,0] == 16:
                    n = min(max(round(0.25*rows_1),8), 27)
                    
                elif feuille_7.iloc[k,0] == 17:
                    n = min(max(round(0.5*rows_1),14), 48)
                    
                # The number of other vehicles is given in the dataset
                else:
                    n = feuille_7.iloc[k,4]
        
            # Else the number of vehicles of a given type is given in the dataset
            else:
                n = feuille_7.iloc[k,4]
        
        for l in range(0, n):
            # Vehicles id should start at 1 (0 being equivalent to vehicle absence)
            if not ID:
                i = 1
            else:
                i = ID[-1] + 1
            ID.append(i)
            type.append(feuille_7.iloc[k,0])

    nb = len(ID)
    
    
    ## Vehicles characteristics
    maxCap = [0 for k in range(0, nb)]
    speed = [0 for k in range(0, nb)]
    tbase = [0 for k in range(0, nb)]
    depot = [[] for k in range(0, nb)]
    services = [[] for k in range(0, nb)]
    fpairings = [[] for k in range(0, nb)]
    i = 0
    
    for l in range(0, nb):
        for k in range(0, rows_7):
            if feuille_7.iloc[k,0] == type[l]:
                speed[l] = feuille_7.iloc[k,12]
                tbase[l] = feuille_7.iloc[k,11]
                #For AC, the MaxCap is equal to the total number of ULDs that the corresponding AC can welcome
                if type[l] == 19: 
                    for j in range(0, rows_6):
                        if AC.type[i] == doc_6.iloc[j,0]:
                            maxCap[l] = doc_6.iloc[j,5]
                    i += 1
                else:
                    maxCap[l] = feuille_7.iloc[k,6]
        
        for x in range(0, cols_8):
            if feuille_8.iloc[0,x] == type[l]:
                for y in range(1, rows_8):
                    if feuille_8.iloc[y,x] != -1 :
                        depot[l].append(feuille_8.iloc[y,x])
        for p in range(0, cols_9):
            for y in range(1, rows_9):
                if feuille_9.iloc[0,p] == type[l]:
                    if feuille_9.iloc[y,p] != -1 :
                        services[l].append(feuille_9.iloc[y,p])
        for t in range(0, cols_10):
            if feuille_10.iloc[0,t] == type[l]:
                for y in range(1, rows_10):
                    if feuille_10.iloc[y,t] != -1 :
                        fpairings[l].append(feuille_10.iloc[y,t])
    
    
    ## Vehicles service time definition
    c = [[] for k in range(0, nb)]
    cD = [[] for k in range(0, nb)]
    v = [[] for k in range(0, nb)]
    vD = [[] for k in range(0, nb)]
    
    feuille_11 = pd.read_excel(doc_7, "dollies_data")
    rows_11 = len(feuille_11)
    feuille_12 = pd.read_excel(doc_7, "HL_data")
    rows_12 = len(feuille_12)
    feuille_13 = pd.read_excel(doc_7, "SL_data")
    rows_13 = len(feuille_13)
    feuille_14 = pd.read_excel(doc_7, "CT_data")
    rows_14 = len(feuille_14)
    
    for l in range(0, nb):
        if len(services[l]) == 1:
            for k in range(0, rows_7):
                if feuille_7.iloc[k,0] == type[l]:
                     c[l].append(feuille_7.iloc[k,7])
                     cD[l].append(feuille_7.iloc[k,8])
                     v[l].append(feuille_7.iloc[k,9])
                     vD[l].append(feuille_7.iloc[k,10])
        elif type[l] == 14:
            for p in range(0, rows_11):
                c[l].append(feuille_11.iloc[p,1])
                cD[l].append(feuille_11.iloc[p,2])
                v[l].append(feuille_11.iloc[p,3])
                vD[l].append(feuille_11.iloc[p,4])
        elif type[l] == 15:
            for p in range(0, rows_12):
                c[l].append(feuille_12.iloc[p,1])
                cD[l].append(feuille_12.iloc[p,2])
                v[l].append(feuille_12.iloc[p,3])
                vD[l].append(feuille_12.iloc[p,4])
        elif type[l] == 16:
            for p in range(0, rows_13):
                c[l].append(feuille_13.iloc[p,1])
                cD[l].append(feuille_13.iloc[p,2])
                v[l].append(feuille_13.iloc[p,3])
                vD[l].append(feuille_13.iloc[p,4])
        elif type[l] == 17:
            for p in range(0, rows_14):
                c[l].append(feuille_14.iloc[p,1])
                cD[l].append(feuille_14.iloc[p,2])
                v[l].append(feuille_14.iloc[p,3])
                vD[l].append(feuille_14.iloc[p,4])
    
    
    ## Vehicles services sequence to follow
    sequence = [[] for k in range(0, nb)]
    
    for l in range(0, nb):
        if type[l] == 14:
            sequence[l].append([6,7,10,11,20])
            sequence[l].append([12,13,14,15,16])
    
    
    ## Vehicles additional characteristics initialization
    # Vehicles current capacities are initialized to their max capacity
    currentCap = [0 for k in range(0, nb)]
    for k in range(0, nb):
        currentCap[k] = maxCap[k]
    
    # Initially, vehicles are not used and therefore not involved in any services
    currentService = ["any" for k in range(0, nb)]
    used = [False for k in range(0, nb)]
    
    # Vehicles are initially available 24h before the beginning of the time horizon
    if min(AC.STA).day == 1: 
        # If the day considered is the 1st of the month, look at the last day of previous month
        available_init = [dt.datetime(min(AC.STA).year, min(AC.STA).month-1, monthrange(min(AC.STA).year, min(AC.STA).month-1)[1], 12, 0, 0) for k in range(0, nb)]
    else:        
        available_init = [dt.datetime(min(AC.STA).year, min(AC.STA).month, min(AC.STA).day-1, 12, 0, 0) for k in range(0, nb)]
    
    # Aircraft are available at their STA (not before)
    j = 0
    for k in range(0, nb):
        if type[k] == 19:
            available_init[k] = AC.STA[j]
            j += 1
            
    ttmp = [dt.datetime(min(AC.STA).year, min(AC.STA).month, min(AC.STA).day, 12, 0, 0) for k in range(0, nb)]
    
    # Vehicles initial list of compatible AC is defined based on forbidden pairings
    compatibleAC_init = [[] for k in range(0, nb)]
    j = 0
    for k in range(0, nb):
        for i in AC.ID:
            #Vehicles are compatible with all aircraft whose type is not in the vehicle forbidden pairings
            if type[k] != 19:
                if AC.type[i] not in fpairings[k]:
                    compatibleAC_init[k].append(i)
        #AC are only compatible with themselves
        if type[k] == 19:
            compatibleAC_init[k].append(j)
            j += 1
    
    currentPos = []
    i = 0
    j = 0
    for k in range(0, nb):
        #Warehouses IN have a fixed position (AC.IN)
        if type[k] == 18:
            currentPos.append(AC.IN[i])
            i+= 1
        #AC have a fixed position (AC.parking)
        elif type[k] == 19:
            currentPos.append(AC.parking[j])
            j += 1
        #Vehicles' initial positions are randomly picked among the vehicles' depots 
        else:
            currentPos.append(depot[k][0]) #if you want a fixed initial depot
            #currentPos.append(random.choice(depot[k])) #if you want to add some randomness
        
    
    ## Authorize subsequent modifications of some variables
    def __init__(self):
        self.currentCap_init = self.currentCap.copy()
        self.currentCap = self.currentCap.copy()
        self.currentService = self.currentService.copy()
        self.used = self.used.copy()
        self.available = self.available_init.copy()
        self.ttmp = self.ttmp.copy()
        self.currentPos = self.currentPos.copy()
        # Create a copy of compatibleAC_init to keep stored the initial list of compatible AC
        self.compatibleAC = self.compatibleAC_init.copy()
        
# Create a class object to use vehicles data in other data classes
V = vehicle()
  

#####  -----------------------  Solution vectors  -----------------------  #####

class solution:
    TripPosition = [[] for k in range(0, V.nb)]
    TripSchedule = [[] for k in range(0, V.nb)]
    EndTime = [[dt.datetime(min(AC.STA).year, min(AC.STA).month, min(AC.STA).day, 12, 0, 0) for s in range (0, len(S.ID))] for i in range(0, len(AC.ID))]

    ## Authorize solution vectors to be completed afterwards
    def __init__(self):
        self.TripPosition = self.TripPosition.copy()
        self.TripSchedule = self.TripSchedule.copy()
        self.EndTime = self.EndTime.copy()




